#!/usr/bin/env perl

use strict;
use warnings;

use File::Basename;
use File::Spec;
use File::Spec::Link;

my $this_script = File::Spec::Link->resolve($0);
$this_script = File::Spec->rel2abs($this_script);
my ($scriptname, $scriptdir) = fileparse($this_script);
$scriptdir = File::Spec->rel2abs($scriptdir);
my $tabix = File::Spec->catfile($scriptdir, 'tabix/tabix');
my $bgzip = File::Spec->catfile($scriptdir, 'tabix/bgzip');
my $version = '1.0.12';
my $ERROR_PREFIX = '[REAPR pipeline]';

if ($#ARGV == -1) {
    die qq/REAPR version: $version
Usage:
    reapr <task> [options]

Common tasks:
    facheck    - checks IDs in fasta file
    perfectmap - make perfect uniquely mapping plot files
    pipeline   - runs the REAPR pipeline, using an assembly and mapped reads
                 as input, and optionally results of perfectmap.
                 (It runs preprocess, stats, fcdrate, score, summary and break)
    plots      - makes Artemis plot files for a given contig, using results
                 from stats (and optionally results from score)

Advanced tasks:
    preprocess - preprocess files: necessary for running stats
    stats      - generates stats from a BAM file
    fcdrate    - estimates FCD cutoff for score, using results from stats
    score      - calculates scores and assembly errors, using results from stats
    summary    - make summary stats file, using results from score
    break      - makes broken assembly, using results from score
    gapresize  - experimental, calculates gap sizes based on read mapping
    perfectfrombam - generate perfect mapping plots from a bam file (alternative
                     to using perfectmap for large genomes)
/;
}


my %tasks = (
    'perfectmap' => "task_perfectmap.pl",
    'preprocess' => "task_preprocess.pl",
    'stats' => "task_stats",
    'break' => "task_break",
    'score' => "task_score",
    'plots' => "task_plots.pl",
    'pipeline'  => "x",
    'summary' => "task_summary.pl",
    'facheck' => 'task_facheck.pl',
    'gapresize' => 'task_gapresize',
    'fcdrate' => 'task_fcdrate',
    'perfectfrombam' => 'task_perfectfrombam.pl',
);

for my $k (keys %tasks) {
    $tasks{$k} = File::Spec->catfile($scriptdir, $tasks{$k});
}


if ($tasks{$ARGV[0]}) {
    if ($ARGV[0] eq "pipeline") {
        if ($#ARGV < 3 or $#ARGV > 4) {
            die qq/usage: reapr pipeline <assembly.fa> <in.bam> <out directory> [perfectmap prefix]

where 'perfectmap prefix' is optional and should be the prefix used when task
perfectmap was run.

It is assumed that reads in in.bam are 'innies', i.e. the correct orientation
is reads in a pair pointing towards each other (---> <---).
/;
        }
        my $ref = $ARGV[1];
        my $bam = $ARGV[2];
        my $dir = $ARGV[3];
        my $stats_prefix = '01.stats';
        my $fcdrate_prefix = '02.fcdrate';
        my $score_prefix = '03.score';
        my $break_prefix = '04.break';
        my $summary_prefix = '05.summary';

        my $perfect_prefix = "";
        if ($#ARGV == 4) {
            $perfect_prefix = File::Spec->rel2abs($ARGV[4]);
        }

        print "$ERROR_PREFIX REAPR version $version.  Running pipeline with:\n\treapr " . join(' ', @ARGV) . "\n\n";

        # run preprocess
        my $cmd = "preprocess $ref $bam $dir";
        print "$ERROR_PREFIX Running preprocess:\n\treapr $cmd\n\n";
        system_call("$this_script $cmd");

        chdir $dir or die "$ERROR_PREFIX Error chdir to '$dir'";
        print "$ERROR_PREFIX cd $dir\n";

        # run stats
        $cmd = "stats";
        $cmd .= " -p $perfect_prefix.perfect_cov.gz" if $perfect_prefix;
        $cmd .= " ./ $stats_prefix";
        print "\n$ERROR_PREFIX Running stats:\n\treapr $cmd\n\n";
        system_call("$this_script $cmd");

        # run fcdrate
        $cmd = "fcdrate ./ $stats_prefix $fcdrate_prefix";
        print "$ERROR_PREFIX Running fcdrate:\n\treapr $cmd\n\n";
        system_call("$this_script $cmd");

        # get fcd cutoff from fcdrate output
        my $fcdrate_info_file = "$fcdrate_prefix.info.txt";
        open F, $fcdrate_info_file or die "$ERROR_PREFIX Error opening file '$fcdrate_info_file'\n";
        <F>;
        my $tmp = <F>;
        my ($fcd_cutoff) = split(/\t/, $tmp);
        close F;

        # run score
        $cmd = "score";
        $cmd .= " -P 5" if $perfect_prefix;
        $cmd .= " 00.assembly.fa.gaps.gz 00.in.bam $stats_prefix $fcd_cutoff $score_prefix";
        print "$ERROR_PREFIX Running score:\n\treapr $cmd\n\n";
        system_call("$this_script $cmd");

        # run break
        $cmd = "break 00.assembly.fa $score_prefix.errors.gff.gz $break_prefix";
        print "$ERROR_PREFIX Running break:\n\treapr $cmd\n\n";
        system_call("$this_script $cmd");

        # run summary
        $cmd = "summary 00.assembly.fa $score_prefix $break_prefix $summary_prefix";
        print "$ERROR_PREFIX Running summary:\n\treapr $cmd\n\n";
        system_call("$this_script $cmd");
    }
    elsif ($#ARGV == 0) {
        print STDERR "usage:\nreapr $ARGV[0] ";
        exec "$tasks{$ARGV[0]} --wrapperhelp" or die;
    }
    else {
        my $cmd;

        if ($ARGV[0] eq "score") {
            my $score_out = "$ARGV[-1].per_base.gz";
            my $errors_out = "$ARGV[-1].errors.gff";
            $cmd = "$tasks{$ARGV[0]} " . join(" ", @ARGV[1..$#ARGV]) . "  | $bgzip -f -c > $score_out && $tabix -f -b 2 -e 2 $score_out && $bgzip -f $errors_out && $tabix -f -p gff $errors_out.gz";
        }
        else {
            $cmd = "$tasks{$ARGV[0]} " . join(" ", @ARGV[1..$#ARGV]);

            if ($ARGV[0] eq "stats") {
                my $outfile = "$ARGV[-1].per_base.gz";
                $cmd .= "  | $bgzip -f -c > $outfile && $tabix -f -b 2 -e 2 $outfile";
            }
        }

        exec $cmd or die;
    }
}
else {
    die qq/Task "$ARGV[0]" not recognised.\n/;
}

# usage: system_call(string)
# Runs the string as a system call, dies if call returns nonzero error code
sub system_call {
    my $cmd  = shift;
    if (system($cmd)) {
        print STDERR "Error in system call:\n$cmd\n";
        exit(1);
    }
}
